library(COMPoissonReg)
Dcmp=Vectorize(dcmp)

# CMP minus log-likelihood: ########################################################################
CMPmllk = function(wpar){
  lambda  = exp(wpar[1])
  nu      = exp(wpar[2])
  -sum(log(Dcmp(x,lambda,nu))) }

# Independent mixture of Poissons, minus log-likelihood: ###########################################
n2w = function(lambda,delta)log(c(lambda,delta[-1]/(1-sum(delta[-1]))))
w2n = function(wpar){
  m       = (length(wpar)+1)/2
  lambda  = exp(wpar[1:m])
  delta   = exp(c(0,wpar[(m+1):(2*m-1)]))
  return(list(lambda=lambda,delta=delta/sum(delta))) }

mllk = function(wpar,x){
  zzz = w2n(wpar)
  -sum(log(outer(x,zzz$lambda,dpois)%*%zzz$delta)) }

#Poisson-HMM: ######################################################################################
### Transforming natural parameters to working
pois.HMM.pn2pw = function(m,lambda,gamma,delta=NULL,stationary=TRUE){
  tlambda = log(lambda)
  if(m==1) return(tlambda)
  foo     = log(gamma/diag(gamma))
  tgamma  = as.vector(foo[!diag(m)])
  if(stationary) {tdelta  = NULL}
  else {tdelta = log(delta[-1]/delta[1])}
  parvect = c(tlambda,tgamma,tdelta)
  return(parvect) }

### Transforming working parameters to natural
pois.HMM.pw2pn = function(m,parvect,stationary=TRUE){
  lambda        = exp(parvect[1:m])
  gamma         = diag(m)
  if (m==1) return(list(lambda=lambda,gamma=gamma,delta=1))
  gamma[!gamma] = exp(parvect[(m+1):(m*m)])
  gamma         = gamma/apply(gamma,1,sum)
  if(stationary){delta=solve(t(diag(m)-gamma+1),rep(1,m))}
  else {foo=c(1,exp(parvect[(m*m+1):(m*m+m-1)]))
  delta         = foo/sum(foo)}
  return(list(lambda=lambda,gamma=gamma,delta=delta)) }

### Computing minus the log-likelihood from the working parameters
pois.HMM.mllk = function(parvect,x,m,stationary=TRUE,...){
  n        = length(x)
  pn       = pois.HMM.pw2pn(m,parvect,stationary=stationary)
  foo      = pn$delta*dpois(x[1],pn$lambda)
  sumfoo   = sum(foo)
  lscale   = log(sumfoo)
  foo      = foo/sumfoo
  for (i in 2:n){
    if(!is.na(x[i])){P=dpois(x[i],pn$lambda)}
    else {P=rep(1,m)}
    foo    = foo %*% pn$gamma*P
    sumfoo = sum(foo)
    lscale = lscale+log(sumfoo)
    foo    = foo/sumfoo }
  mllk     = -lscale
  return(mllk) }

### Computing the MLEs, given starting values for the natural parameters
pois.HMM.mle = function(x,m,lambda0,gamma0,delta0=NULL,stationary=TRUE,...){
  parvect0  = pois.HMM.pn2pw(m,lambda0,gamma0,delta0,stationary=stationary)
  mod       = nlm(pois.HMM.mllk,parvect0,x=x,m=m,stationary=stationary,iterlim=1000)
  pn        = pois.HMM.pw2pn(m=m,mod$estimate,stationary=stationary)
  mllk      = mod$minimum
  np        = length(parvect0)
  AIC       = 2*(mllk+np)
  n         = sum(!is.na(x))
  BIC       = 2*mllk+np*log(n)
  list(m=m,lambda=pn$lambda,gamma=pn$gamma,delta=pn$delta,code=mod$code,mllk=mllk,AIC=AIC,BIC=BIC) }

#CMP-HMM: ##########################################################################################
### Transforming natural parameters to working
CMP.HMM.pn2pw = function(m,lambda,nu,gamma,delta=NULL){
  tlambda = log(lambda)
  tnu     = log(nu)
  foo     = log(gamma/diag(gamma))
  tgamma  = as.vector(foo[!diag(m)])
  tdelta  = NULL
  parvect = c(tlambda,tnu,tgamma,tdelta)
  return(parvect) }

### Transforming working parameters to natural
CMP.HMM.pw2pn = function(m,parvect){
  lambda        = exp(parvect[1:m])
  nu            = exp(parvect[(m+1):(2*m)])
  gamma         = diag(m)
  gamma[!gamma] = exp(parvect[(2*m+1):(m*m+m)])
  gamma         = gamma/apply(gamma,1,sum)
  delta         = solve(t(diag(m)-gamma+1),rep(1,m))
  return(list(lambda=lambda,nu=nu,gamma=gamma,delta=delta)) }

### Computing minus the log-likelihood from the working parameters
CMP.HMM.mllk = function(parvect,x,m,...){
  n        = length(x)
  pn       = CMP.HMM.pw2pn(m,parvect)
  foo      = pn$delta*Dcmp(x[1],pn$lambda,pn$nu)
  sumfoo   = sum(foo)
  lscale   = log(sumfoo)
  foo      = foo/sumfoo
  for (i in 2:n){  
    if(!is.na(x[i])){P=Dcmp(x[i],pn$lambda,pn$nu)}
    else {P=rep(1,m)}
    foo    = foo %*% pn$gamma*P
    sumfoo = sum(foo)
    lscale = lscale+log(sumfoo)
    foo    = foo/sumfoo }
  mllk     = -lscale
  return(mllk) }

### Computing minus the log-likelihood from the natural parameters
CMP.HMM.mllk_np = function(parvect,x,m,...){
  n             = length(x)
  lambda        = parvect[1:m]
  nu            = parvect[(m+1):(2*m)]
  gamma         = diag(m)
  gamma[!gamma] = parvect[(2*m+1):(m*m+m)]
  gamma         = t(gamma)
  for (i in 1:m) {gamma[i,i] = 1-sum(gamma[i,-i])}
  delta         = solve(t(diag(m)-gamma+1),rep(1,m))
  foo           = delta*Dcmp(x[1],lambda,nu)
  sumfoo        = sum(foo)
  lscale        = log(sumfoo)
  foo           = foo/sumfoo
  for (i in 2:n){  
    if(!is.na(x[i])){P=Dcmp(x[i],lambda,nu)}
    else {P=rep(1,m)}
    foo     = foo %*% gamma*P
    sumfoo  = sum(foo)
    lscale  = lscale+log(sumfoo)
    foo     = foo/sumfoo }
  mllk          = -lscale
  return(mllk) }

### Computing the MLEs, given starting values for the natural parameters
CMP.HMM.mle = function(x,m,lambda0,nu0,gamma0,...){
  parvect0  = CMP.HMM.pn2pw(m,lambda0,nu0,gamma0)
  np        = length(parvect0)
  mod       = nlm(CMP.HMM.mllk,parvect0,x=x,m=m,print.level=1,stepmax=1,iterlim=1000,hessian=T)
  pn        = CMP.HMM.pw2pn(m=m,mod$estimate)
  mllk      = mod$minimum
  AIC       = 2*(mllk+np)
  n         = sum(!is.na(x))
  BIC       = 2*mllk+np*log(n)
  list(m=m,lambda=pn$lambda,nu=pn$nu,gamma=pn$gamma,delta=pn$delta,code=mod$code,hessian=mod$hessian,
       mllk=mllk,AIC=AIC,BIC=BIC) }

###  Functions needed to find standard errors:
list_params = function(m,parvect){
  lambda        = parvect[1:m]
  nu            = parvect[(m+1):(2*m)]
  gamma         = diag(m)
  gamma[!gamma] = parvect[(2*m+1):(m*m+m)]
  gamma         = gamma/apply(gamma,1,sum)
  delta         = solve(t(diag(m)-gamma+1),rep(1,m))
  return(list(lambda=lambda,nu=nu,gamma=gamma,delta=delta)) }

CMP.HMM.mle_np = function(x,m,lambda0,nu0,gamma0...){
  parvect0  = c(lambda0,nu0,gamma0,delta0 = NULL)
  np        = length(parvect0)
  mod       = nlm(CMP.HMM.mllk_np,parvect0,x=x,m=m,print.level=1,stepmax=1,iterlim=1000,hessian=T)
  pn        = list_params(m,mod$estimate)
  mllk      = mod$minimum
  AIC       = 2*(mllk+np)
  n         = sum(!is.na(x))
  BIC       = 2*mllk+np*log(n)
  list(m=m,lambda=pn$lambda,nu=pn$nu,gamma=pn$gamma,delta=pn$delta,code=mod$code,hessian=mod$hessian,
       mllk=mllk,AIC=AIC,BIC=BIC) }

### Functions needed to compute the mean and variance of a CMP distribution
Z = function(lambda,nu,upper){
  factor = 1
  Z      = 1
  for(y in 1:upper) {factor=factor*lambda/y^nu; Z=Z+factor}
  Z }

means = function(lambda,nu,upper){
  ZZ     = Z(lambda,nu,upper)
  factor = 1/ZZ
  sum    = 0
  for(y in 1:upper) {factor=factor*lambda/y^nu; sum=sum+y*factor} 
  sum }

variances = function(lambda,nu,upper){
  ZZ     = Z(lambda,nu,upper)
  mu     = means(lambda,nu,upper)
  factor = 1/ZZ
  sum    = (0-mu)^2*factor
  for(y in 1:upper) {factor=factor*lambda/y^nu; sum=sum+(y-mu)^2*factor} 
  sum }

means = Vectorize(means)
variances = Vectorize(variances)

# Function to display and compare results:
compare_res = function(x,model,...){
  deltas    = model$delta
  lambdas   = model$lambda
  nus       = model$nu
  Gamma     = model$gamma
  mus       = means(lambdas,nus,10000)
  vars      = variances(lambdas,nus,10000)
  modmean   = sum(deltas*mus)
  samp_mean = mean(x,na.rm=T)
  excess    = deltas[1]*deltas[2]*(mus[1]-mus[2])^2  # Valid for 2-state only
  modvar    = sum(deltas*vars)+excess
  samp_var  = var(x,na.rm=T) 
  A         = excess/(excess+sum(deltas*vars)) 
  w         = 1-Gamma[1,2]-Gamma[2,1] 
  kk        = 1:25
  modelACF  = A*w^kk
  round(modelACF,4)
  list(model_mean=modmean,sample_mean=samp_mean,model_var=modvar,sample_var=samp_var,
       model_ACF=modelACF,A=A,w=w) }

save.image("AllFunctionsNeeded.RData")