# Fit m-state CMPHMM to the selected data set. 

load("AllFunctionsNeeded.RData")

x=scan("Fuerth_pedestrians_505.txt")
x=scan("Weiss_downloads_241.txt")
x=scan("Weiss_emergencies_96.txt")
x=scan("Jung_gold_380.txt")     # Jung version of gold particles. (380)
x=scan("Guttorp_gold_1598.txt") # Guttorp v. of gold particles.  (1598)

# Fit 2-state CMP-HMM:
m=2
(lambda0=c(1.5, 3.9))
(nu0=c(2.1,1.4))
(gamma0=matrix(
    c(0.97, 0.03,
      0.04, 0.96),
    m,m,byrow=TRUE))
(mod2s=CMP.HMM.mle(x,m,lambda0,nu0,gamma0))
(result=compare_res(x,mod2s))

# To find std errors:
# Starting parameters are those found from previous optimization multiplied by 'mult'
mult=0.999
lambda0=mult*mod2s$lambda
nu0=mult*mod2s$nu
gamma0=mult*c(mod2s$gamma[1,2],mod2s$gamma[2,1])
(mod2_np=CMP.HMM.mle_np(x,m,lambda0,nu0,gamma0))
(std_errors=sqrt(diag(solve(mod2_np$hessian))))

#############################################################################################
# Now consider without imputation - for Guttorp_gold_1598 only
xx=scan("Guttorp_gold_1598.txt") # Guttorp v. of gold particles.  (1598)
missing=c(123, 238, 423, 479, 559, 580, 601, 657, 991, 1174, 1410)
xx[missing]=NA

# Fit 2-state CMP-HMM:
m=2
(lambda0=c(1.5, 3.9))
(nu0=c(2.1,1.4))
(gamma0=matrix(
    c(0.97, 0.03,
      0.04, 0.96),
    m,m,byrow=TRUE))
(mod2s=CMP.HMM.mle(xx,m,lambda0,nu0,gamma0))
(result=compare_res(xx,mod2s))

# To find std errors:
# Starting parameters are those found from previous optimization multiplied by 'mult'
mult=0.999
lambda0=mult*mod2s$lambda
nu0=mult*mod2s$nu
gamma0=mult*c(mod2s$gamma[1,2],mod2s$gamma[2,1])
(mod2_np=CMP.HMM.mle_np(xx,m,lambda0,nu0,gamma0))
(std_errors=sqrt(diag(solve(mod2_np$hessian))))

#############################################################################################
# Other models:
# 3-state CMP-HMM:
m=3
lambda0=c(1.5,3.9,1.5) 
(gamma0=matrix(
    c(0.9, 0.05, 0.05,
      0.05, 0.9, 0.05,
      0.05, 0.05, 0.9),
    m,m,byrow=TRUE))
(mod3s=CMP.HMM.mle(x,m,lambda0,rep(1,m),gamma0)) 

deltas=mod3s$delta
mus=means(mod3s$lambda,mod3s$nu,10000)
vars=variances(mod3s$lambda,mod3s$nu,10000)
(modmean=sum(deltas*mus))
excess=0
for(i in 1:(m-1)){ for(j in (i+1):m){excess=excess+deltas[i]*deltas[j]*(mus[i]-mus[j])^2} }
(modvar=sum(deltas*vars)+excess)

# 4-state CMP-HMM:
m=4
(lambda0=c(1.5,3.9,1.5,1))
(gamma0=matrix(
  c(0.9, 0.05, 0.03, 0.02,
    0.03, 0.9, 0.05, 0.02,
    0.05, 0.03, 0.9, 0.02, 
    0.05, 0.03, 0.02, 0.9),
  m,m,byrow=TRUE))
(mod4s=CMP.HMM.mle(x,m,lambda0,rep(1,m),gamma0)) 

deltas=mod4s$delta
mus=means(mod4s$lambda,mod4s$nu,10000)
vars=variances(mod4s$lambda,mod4s$nu,10000)
(modmean=sum(deltas*mus))
excess=0
for(i in 1:(m-1)){ for(j in (i+1):m){excess=excess+deltas[i]*deltas[j]*(mus[i]-mus[j])^2} }
(modvar=sum(deltas*vars)+excess)

# Single Poisson:
(minusl=-sum(dpois(x,mean(x),log=T)))
(AIC=2*(minusl+1))
(BIC=2*minusl+1*log(length(x)))

# Single CMP:
start=log(c(1.5,1))
CMPmllk(start)
(CMPmodel=nlm(CMPmllk,start,stepmax=1))
exp(CMPmodel$estimate)
(AIC=2*(CMPmodel$minimum+2))
(BIC=2*(CMPmodel$minimum)+2*log(length(x)))

# Mixture of 2 Poissons:
m=2
wpar=n2w(c(0.5,4),c(1,1)/2)
model=nlm(mllk,wpar,x,print.level=1)
(model2P=w2n(model$estimate))
(modelmean=sum(model2P$delta*model2P$lambda))
(modelvar=modelmean+model2P$delta[1]*model2P$delta[2]*(model2P$lambda[1]-model2P$lambda[2])^2)
c(mean(x),var(x))
(AIC=2*(model$minimum+3))
(BIC=2*(model$minimum)+3*log(length(x)))

# 2-state Poisson-HMM
m=2
lambda0=c(1,2)
gamma0=matrix(
  c(0.9,0.1,
    0.1,0.9),
  m,m,byrow=TRUE)
(mod2s=pois.HMM.mle(x,m,lambda0,gamma0,stationary=TRUE))
(mod_mean=sum(mod2s$lambda*mod2s$delta))
(mod_var=sum(mod2s$lambda*mod2s$delta)+mod2s$delta[1]*mod2s$delta[2]*(mod2s$lambda[1]-
                                                                        mod2s$lambda[2])^2)